#ifndef TREE_SITTER_WASM_INTTYPES_H_
#define TREE_SITTER_WASM_INTTYPES_H_

// https://github.com/llvm/llvm-project/blob/0c3cf200f5b918fb5c1114e9f1764c2d54d1779b/libc/include/llvm-libc-macros/inttypes-macros.h#L209

#define PRId32 "d"

#endif // TREE_SITTER_WASM_INTTYPES_H_
